import abc
import datetime as dt
import os
from functools import cached_property
from logging import getLogger

import polars as pl

from bayesline.api._src.equity.portfoliohierarchy_settings import (
    PortfolioHierarchySettings,
)
from bayesline.api._src.equity.report_settings import (
    PortfolioReportAccessorSettings,
    ReportSettings,
    ReportSettingsMenu,
)
from bayesline.api._src.registry import (
    AsyncReadOnlyRegistry,
    AsyncRegistryBasedApi,
    ReadOnlyRegistry,
    RegistryBasedApi,
)
from bayesline.api._src.types import DateLike, DNFFilterExpressions

logger = getLogger(__name__)


class IllegalPathError(Exception):
    pass


class BasePortfolioReportAccessor(abc.ABC):

    @property
    @abc.abstractmethod
    def axes(self) -> dict[str, list[str]]: ...

    @property
    @abc.abstractmethod
    def metric_cols(self) -> list[str]: ...

    @property
    @abc.abstractmethod
    def pivot_cols(self) -> list[str]: ...

    @cached_property
    def axis_lookup(self) -> dict[str, str]:
        return {
            level: dimension
            for dimension, levels in self.axes.items()
            for level in levels
        }

    def is_path_valid(
        self,
        path_levels: list[str],
        *,
        expand: tuple[str, ...] = (),
    ) -> bool:
        try:
            self.validate_path(path_levels, expand)
            return True
        except AssertionError:
            return False

    def _validate_path(self, path_levels: list[str]) -> list[str]:
        msgs = []
        unknown_levels = [
            level for level in path_levels if level not in self.axis_lookup
        ]
        if unknown_levels:
            msgs.append(f"Unknown levels: {unknown_levels}")

        seen_axes = set()
        prev_axis = ""
        for level in path_levels:
            axis = self.axis_lookup.get(level)
            if not axis:
                msgs.append(f"Level {level} does not exist")
                break

            if axis != prev_axis and axis in seen_axes:
                msgs.append(
                    "Mixed axis groups: "
                    f"{', '.join([self.axis_lookup[level] for level in path_levels])}",
                )
                break

            seen_axes.add(axis)
            prev_axis = axis

        # check that for each dimension the levels are in the right order
        for dimension, levels in self.axes.items():
            path_levels_for_dimension = [
                level for level in path_levels if self.axis_lookup[level] == dimension
            ]
            correct_order = [
                level for level in levels if level in path_levels_for_dimension
            ]
            if path_levels_for_dimension != correct_order:
                msgs.append(
                    f"Invalid order for {dimension}: "
                    f"{', '.join(path_levels_for_dimension)} "
                    f"should be in {', '.join(correct_order)}",
                )

        return msgs

    def validate_path(
        self,
        path_levels: list[str],
        expand: tuple[str, ...] = (),
    ) -> None:
        self._validate_path(path_levels)
        msgs = []
        seen = set()
        for e in expand:
            if e in seen:
                msgs.append(f"Duplicate expand: {e}")
            msgs.extend(self._validate_path([*path_levels, e]))
            seen.add(e)

        if msgs:
            raise IllegalPathError(os.linesep.join(msgs))


class PortfolioReportAccessor(BasePortfolioReportAccessor):

    @abc.abstractmethod
    def get_level_values(
        self,
        levels: tuple[str, ...] = (),
        include_totals: bool = False,
        filters: DNFFilterExpressions | None = None,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    def get_data(
        self,
        path: list[tuple[str, str]],
        *,
        expand: tuple[str, ...] = (),
        pivot_cols: tuple[str, ...] = (),
        value_cols: tuple[str, ...] = (),
        filters: DNFFilterExpressions | None = None,
        pivot_total: bool = False,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    def persist(self, name: str) -> int:
        """
        Persists the given report for the given name.

        Parameters
        ----------
        name : str
            The name to persist the report under. Will throw if the name already exists.

        Returns
        -------
        int
            A globally unique identifier of the persisted report.
        """


class AsyncPortfolioReportAccessor(BasePortfolioReportAccessor):

    @abc.abstractmethod
    async def get_level_values(
        self,
        levels: tuple[str, ...] = (),
        include_totals: bool = False,
        filters: DNFFilterExpressions | None = None,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def get_data(
        self,
        path: list[tuple[str, str]],
        *,
        expand: tuple[str, ...] = (),
        pivot_cols: tuple[str, ...] = (),
        value_cols: tuple[str, ...] = (),
        filters: DNFFilterExpressions | None = None,
        pivot_total: bool = False,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def persist(self, name: str) -> int:
        """
        Persists the given report for the given name.

        Parameters
        ----------
        name : str
            The name to persist the report under. Will throw if the name already exists.

        Returns
        -------
        int
            A globally unique identifier of the persisted report.
        """


class ReportPersister(abc.ABC):

    @abc.abstractmethod
    def persist(
        self,
        name: str,
        settings: PortfolioReportAccessorSettings,
        accessors: list[PortfolioReportAccessor],
    ) -> int: ...

    @abc.abstractmethod
    def load_persisted(self, name_or_id: str | int) -> AsyncPortfolioReportAccessor: ...

    @abc.abstractmethod
    def delete_persisted(self, name_or_id: list[str | int]) -> None: ...


class AsyncReportPersister(abc.ABC):

    @abc.abstractmethod
    async def persist(
        self,
        name: str,
        settings: PortfolioReportAccessorSettings,
        accessors: list[AsyncPortfolioReportAccessor],
    ) -> int: ...

    @abc.abstractmethod
    async def load_persisted(
        self, name_or_id: str | int
    ) -> AsyncPortfolioReportAccessor: ...

    @abc.abstractmethod
    async def delete_persisted(self, name_or_id: list[str | int]) -> None: ...


class PortfolioReportApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> ReportSettings:
        """
        Returns
        -------
        The settings used to create this report.
        """
        ...

    @abc.abstractmethod
    def dates(self) -> list[dt.date]: ...

    @abc.abstractmethod
    def get_report(
        self,
        order: dict[str, list[str]],
        *,
        date: DateLike | None = None,
        date_start: DateLike | None = None,
        date_end: DateLike | None = None,
        subtotals: list[str] | None = None,
    ) -> PortfolioReportAccessor: ...


class AsyncPortfolioReportApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> ReportSettings:
        """
        Returns
        -------
        The settings used to create this report.
        """
        ...

    @abc.abstractmethod
    async def dates(self) -> list[dt.date]: ...

    @abc.abstractmethod
    async def get_report(
        self,
        order: dict[str, list[str]],
        *,
        date: DateLike | None = None,
        date_start: DateLike | None = None,
        date_end: DateLike | None = None,
        subtotals: list[str] | None = None,
    ) -> AsyncPortfolioReportAccessor: ...


class BayeslinePortfolioReportApi(
    RegistryBasedApi[ReportSettings, ReportSettingsMenu, PortfolioReportApi],
):

    @abc.abstractmethod
    def load(
        self,
        ref_or_settings: str | int | ReportSettings,
        *,
        hierarchy_ref_or_settings: str | int | PortfolioHierarchySettings | None = None,
    ) -> PortfolioReportApi: ...

    @property
    @abc.abstractmethod
    def persisted_report_settings(
        self,
    ) -> ReadOnlyRegistry[PortfolioReportAccessorSettings]: ...

    @abc.abstractmethod
    def load_persisted(self, name_or_id: str | int) -> PortfolioReportAccessor: ...

    @abc.abstractmethod
    def delete_persisted(self, name_or_id: list[str | int]) -> None: ...


class AsyncBayeslinePortfolioReportApi(
    AsyncRegistryBasedApi[ReportSettings, ReportSettingsMenu, AsyncPortfolioReportApi],
):

    @abc.abstractmethod
    async def load(
        self,
        ref_or_settings: str | int | ReportSettings,
        *,
        hierarchy_ref_or_settings: str | int | PortfolioHierarchySettings | None = None,
    ) -> AsyncPortfolioReportApi: ...

    @property
    @abc.abstractmethod
    def persisted_report_settings(
        self,
    ) -> AsyncReadOnlyRegistry[PortfolioReportAccessorSettings]: ...

    @abc.abstractmethod
    async def load_persisted(
        self, name_or_id: str | int
    ) -> AsyncPortfolioReportAccessor: ...

    @abc.abstractmethod
    async def delete_persisted(self, name_or_id: list[str | int]) -> None: ...



import abc
from typing import Literal

import polars as pl

FactorType = Literal["Market", "Style", "Industry", "Region"]


class BayeslineEquityIdApi:

    @abc.abstractmethod
    def lookup_ids(self, ids: list[str], top_n: int = 0) -> pl.DataFrame:
        """
        Parameters
        ----------
        ids: list[str]
            The ids to lookup.
        top_n: int
            The number of results to return, where `0` denotes all records.

        Returns
        -------
        pl.DataFrame
            a dataframe with all identifiers that could be matched, sorted by
            `id` and `start_date`.
        """
        ...


class AsyncBayeslineEquityIdApi:

    @abc.abstractmethod
    async def lookup_ids(self, ids: list[str], top_n: int = 0) -> pl.DataFrame:
        """
        Parameters
        ----------
        ids: list[str]
            The ids to lookup.
        top_n: int
            The number of results to return, where `0` denotes all records.

        Returns
        -------
        pl.DataFrame
            a dataframe with all identifiers that could be matched, sorted by
            `id` and `start_date`.
        """
        ...



import abc
import datetime as dt
from typing import Literal

import polars as pl

from bayesline.api._src.equity.universe_settings import (
    UniverseSettings,
    UniverseSettingsMenu,
)
from bayesline.api._src.registry import AsyncRegistryBasedApi, RegistryBasedApi
from bayesline.api._src.types import DateLike, IdType


class AssetUniverseApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> UniverseSettings:
        """
        Returns
        -------
        The settings used to create this universe.
        """
        ...

    @property
    @abc.abstractmethod
    def id_types(self) -> list[IdType]:
        """
        Returns
        -------
        supported id types for this universe.
        """
        ...

    @abc.abstractmethod
    def coverage(self, id_type: IdType | None = None) -> list[str]:
        """
        Parameters
        ----------
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`

        Raises
        ------
        ValueError
            If the given id type is not supported.

        Returns
        -------
        list of all asset ids this universe covers, in given id type.
        """
        ...

    @abc.abstractmethod
    def dates(
        self, *, range_only: bool = False, trade_only: bool = False
    ) -> list[dt.date]:
        """
        Parameters
        ----------
        range_only: bool, default=False
            If True, returns the first and last date only.
        trade_only: bool, default=False
            If True, filter down the dats to trade dates only.

        Returns
        -------
        list of all dates this universe covers.
        """

    @abc.abstractmethod
    def counts(
        self,
        dates: bool = True,
        industry_level: int = 0,
        region_level: int = 0,
        universe_type: Literal["estimation", "coverage", "both"] = "both",
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        dates: bool, optional
            If True, groups by dates
        industry_level: int, optional
            The level of industry aggregation to group by.
            0 means no industry aggregation, 1 means level 1, etc.
            Values greater than the max level are treated as the max level.
        region_level: int, optional
            The level of region aggregation to group by.
            0 means no region aggregation, 1 means level 1, etc.
            Values greater than the max level are treated as the max level.
        universe_type: Literal["estimation", "coverage", "both"], optional
            The type of universe to calculate the counts for.
        id_type: IdType, optional
            The id type to calculate the daily stats for, e.g. `ticker`,
            which is relevant as the coverage may differ by id type.
            The given id type must be supported, i.e. in `id_types`.

        Returns
        -------
        pl.DataFrame
            Universe counts.
            If grouped by dates then the count will be given.
            If not grouped by dates then the mean/min/max across
            all dates will be given.
        """
        ...

    @abc.abstractmethod
    def input_id_mapping(
        self,
        *,
        id_type: IdType | None = None,
        filter_mode: Literal["all", "mapped", "unmapped"] = "all",
        mode: Literal[
            "all", "daily-counts", "input-asset-counts", "latest-name"
        ] = "all",
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`, or the default
            ID type of the universe if `None`.
        filter_mode: Literal[all, mapped, unmapped]
            if `mapped` will only consider assets that could be mapped.
            if `unmapped` will only consider assets that could not be mapped.
        mode: Literal[all, daily-counts, latest-name]
            if `all`, returns all dated mappings
            if `daily-counts`, returns the daily counts of mapped assets
            if `input-asset-counts`, returns the total counts of input assets
            if `latest-name`, returns the latest name of mapped assets

        Returns
        -------
        pl.DataFrame
            If mode is `all`, a DataFrame with `date`, `input_asset_id`,
            `input_asset_id_type`, `output_asset_id`, `output_asset_id_type` and,
            `name` columns.
            It contains contains the original input ID space and the mapped ids.
            The mapped IDs will be `None` if for the given date and input ID the
            asset cannot be mapped.
            If mode is `daily-counts`, a DataFrame with `date`, `asset_id` and `count`
            columns.
            If mode is `input-asset-counts`, a DataFrame with `input_asset_id` and `count`
            columns.
            If mode is `latest-name`, a DataFrame with `asset_id` and `name` columns.
        """
        ...

    @abc.abstractmethod
    def get(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        filter_tradedays: bool = False,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        start: DateLike, optional
            The start date of the universe to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported, i.e. in `id_types`.
        filter_tradedays: bool, default=False
            If True, filter down the data to trade dates only.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range.
        """
        ...


class AsyncAssetUniverseApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> UniverseSettings:
        """
        Returns
        -------
        The settings used to create this universe.
        """
        ...

    @property
    @abc.abstractmethod
    def id_types(self) -> list[IdType]:
        """
        Returns
        -------
        supported id types for this universe.
        """
        ...

    @abc.abstractmethod
    async def coverage(self, id_type: IdType | None = None) -> list[str]:
        """
        Parameters
        ----------
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`

        Raises
        ------
        ValueError
            If the given id type is not supported.

        Returns
        -------
        list of all asset ids this universe covers, in given id type.
        """
        ...

    @abc.abstractmethod
    async def dates(
        self, *, range_only: bool = False, trade_only: bool = False
    ) -> list[dt.date]:
        """
        Parameters
        ----------
        range_only: bool, default=False
            If True, returns the first and last date only.
        trade_only: bool, default=False
            If True, filter down the dats to trade dates only.

        Returns
        -------
        list of all dates this universe covers.
        """

    @abc.abstractmethod
    async def counts(
        self,
        dates: bool = True,
        industry_level: int = 0,
        region_level: int = 0,
        universe_type: Literal["estimation", "coverage", "both"] = "both",
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        dates: bool, optional
            If True, groups by dates
        industry_level: int, optional
            The level of industry aggregation to group by.
            0 means no industry aggregation, 1 means level 1, etc.
            Values greater than the max level are treated as the max level.
        region_level: int, optional
            The level of region aggregation to group by.
            0 means no region aggregation, 1 means level 1, etc.
            Values greater than the max level are treated as the max level.
        universe_type: Literal["estimation", "coverage", "both"], optional
            The type of universe to calculate the counts for.
        id_type: IdType, optional
            The id type to calculate the daily stats for, e.g. `ticker`,
            which is relevant as the coverage may differ by id type.
            The given id type must be supported, i.e. in `id_types`.

        Returns
        -------
        pl.DataFrame
            Universe counts.
            If grouped by dates then the count will be given.
            If not grouped by dates then the mean/min/max across
            all dates will be given.
        """
        ...

    @abc.abstractmethod
    async def input_id_mapping(
        self,
        *,
        id_type: IdType | None = None,
        filter_mode: Literal["all", "mapped", "unmapped"] = "all",
        mode: Literal[
            "all", "daily-counts", "input-asset-counts", "latest-name"
        ] = "all",
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`, or the default
            ID type of the universe if `None`.
        filter_mode: Literal[all, mapped, unmapped]
            if `mapped` will only consider assets that could be mapped.
            if `unmapped` will only consider assets that could not be mapped.
        mode: Literal[all, daily-counts, latest-name]
            if `all`, returns all dated mappings
            if `daily-counts`, returns the daily counts of mapped assets
            if `input-asset-counts`, returns the total counts of input assets
            if `latest-name`, returns the latest name of mapped assets

        Returns
        -------
        pl.DataFrame
            If mode is `all`, a DataFrame with `date`, `input_asset_id`,
            `input_asset_id_type`, `output_asset_id`, `output_asset_id_type` and,
            `name` columns.
            It contains contains the original input ID space and the mapped ids.
            The mapped IDs will be `None` if for the given date and input ID the
            asset cannot be mapped.
            If mode is `daily-counts`, a DataFrame with `date` and `count`
            columns.
            If mode is `asset-counts`, a DataFrame with `input_asset_id` and `count`
            columns.
            If mode is `latest-name`, a DataFrame with `asset_id` and `name` columns.
        """
        ...

    @abc.abstractmethod
    async def get(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        filter_tradedays: bool = False,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        start: DateLike, optional
            The start date of the universe to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported, i.e. in `id_types`.
        filter_tradedays: bool, default=False
            If True, filter down the data to trade dates only.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range.
        """
        ...


class BayeslineEquityUniverseApi(
    RegistryBasedApi[UniverseSettings, UniverseSettingsMenu, AssetUniverseApi],
): ...


class AsyncBayeslineEquityUniverseApi(
    AsyncRegistryBasedApi[
        UniverseSettings, UniverseSettingsMenu, AsyncAssetUniverseApi
    ],
): ...



from collections import Counter
from itertools import pairwise
from typing import Annotated

import polars as pl
from pydantic import Field, field_validator, model_validator

from bayesline.api._src.equity.portfolio_settings import PortfolioOrganizerSettings
from bayesline.api._src.registry import Settings, SettingsMenu, SettingsTypeMetaData


class PortfolioHierarchySettings(Settings):
    """
    Specifies portfolio hierarchies with arbitrary groupings (e.g. manager, etc.).
    """

    portfolio_schema: Annotated[
        str | int | PortfolioOrganizerSettings,
        Field(
            description=(
                "The portfolio organizer settings to use as an underlying schema of "
                "portfolios. The 'Default' schema is used by default."
            ),
        ),
        SettingsTypeMetaData[str | int | PortfolioOrganizerSettings](
            references=PortfolioOrganizerSettings
        ),
    ]

    groupings: dict[str, list[str]] = Field(default_factory=dict)
    portfolio_ids: list[str]
    benchmark_ids: list[str | None]

    @classmethod
    def from_polars(
        cls: type["PortfolioHierarchySettings"],
        df: pl.DataFrame,
        portfolio_schema: str | int | PortfolioOrganizerSettings = "Default",
    ) -> "PortfolioHierarchySettings":
        """
        Creates a portfolio hierarchy from a dataframe.

        Must contain a column `portfolio_id` and optionally `benchmark_id`.
        Every other column is interpreted as a grouping. 0 groupings are allowed.
        Index is ignored.

        Parameters
        -----------
            portfolio_schema: str:
                The underlying portfolio schema to use.
            df: pl.DataFrame:
                The dataframe to create the hierarchy from.

        Returns
        -------
        Pydantic object representing the hierarchy.
        """
        if "portfolio_id" not in df.columns:
            raise ValueError("portfolio_id column not found in the dataframe.")

        groupings = {
            col: df.get_column(col).to_list()
            for col in df.columns
            if col not in {"portfolio_id", "benchmark_id"}
        }
        portfolio_ids = df.get_column("portfolio_id").to_list()

        if "benchmark_id" not in df.columns:
            benchmark_ids = [None] * len(portfolio_ids)
        else:
            benchmark_ids = df.get_column("benchmark_id").to_list()

        return cls(
            portfolio_schema=portfolio_schema,
            groupings=groupings,
            portfolio_ids=portfolio_ids,
            benchmark_ids=benchmark_ids,
        )

    def to_polars(self) -> pl.DataFrame:
        """
        Converts the hierarchy to a polars dataframe.

        The last two columns are the portfolio and benchmark IDs.
        Every column before that is a grouping. 0 groupings are possible.

        Returns
        -------
        pl.DataFrame:
            The dataframe representation of the hierarchy, sorted by portfolio ID.
        """
        return pl.DataFrame(self.groupings).with_columns(
            pl.Series("portfolio_id", self.portfolio_ids),
            pl.Series("benchmark_id", self.benchmark_ids),
        )

    @field_validator("portfolio_ids")
    @classmethod
    def _validate_portfolio_ids(
        cls: type["PortfolioHierarchySettings"], v: list[str]
    ) -> list[str]:
        if not v:
            raise ValueError("Portfolio IDs must be non-empty.")

        duplicates = [item for item, count in Counter(v).items() if count > 1]
        if duplicates:
            raise ValueError(
                "Portfolio IDs must be unique. "
                f"Found duplicates: {', '.join(duplicates)}"
            )

        if any(p is None or p.strip() == "" for p in v):
            raise ValueError("Portfolio IDs must be non-empty strings.")

        return v

    @field_validator("portfolio_ids")
    @classmethod
    def _validate_benchmark_ids(
        cls: type["PortfolioHierarchySettings"], v: list[str]
    ) -> list[str]:
        if not v:
            raise ValueError("Benchmark IDs must be non-empty.")
        return v

    @field_validator("groupings")
    @classmethod
    def _validate_groupings(
        cls: type["PortfolioHierarchySettings"], v: dict[str, list[str]]
    ) -> dict[str, list[str]]:
        if not all(len(v1) == len(v2) for v1, v2 in pairwise(v.values())):
            raise ValueError(f"Groupings must have the same length. {v}")

        groups_with_nulls = []
        for group, values in v.items():
            if any(e is None for e in values):
                groups_with_nulls.append(group)
        if groups_with_nulls:
            raise ValueError(
                f"Groupings must not contain null values. Found in: {groups_with_nulls}"
            )

        return v

    @model_validator(mode="after")
    def _validate_dimensions(self) -> "PortfolioHierarchySettings":
        if len(self.portfolio_ids) != len(self.benchmark_ids):
            raise ValueError(
                "Portfolio IDs and benchmark IDs must have the same length."
            )

        n = len(self.portfolio_ids)
        mismatches = [g for g in self.groupings.values() if len(g) != n]
        if len(mismatches) > 0:
            raise ValueError(
                "Portfolio IDs, benchmark IDs and groups must have the same length. "
                f"Found mismatches: {mismatches}"
            )
        return self


class PortfolioHierarchySettingsMenu(
    SettingsMenu[PortfolioHierarchySettings], frozen=True, extra="forbid"
):
    """
    Specifies the set of available portfolios that can be used to create hierarchies.
    """

    schemas: dict[str, list[str]] = Field(
        default_factory=dict,
        description=(
            "Set of available portfolio organizer schemas and their "
            "associated portfolio IDs."
        ),
    )

    def describe(self, settings: PortfolioHierarchySettings | None = None) -> str:
        if settings is None:
            schemas = {k: sorted(v) for k, v in self.schemas.items()}
            return f"Available Portfolio Schemas: {schemas}"
        else:
            schema = settings.portfolio_schema
            if isinstance(schema, str) and schema not in self.schemas:
                return f"Unknown schema: {schema}"

            df = settings.to_polars()
            if isinstance(schema, str):
                df = df.with_columns(
                    pl.col("portfolio_id").map_elements(
                        lambda x: f"[{x}]" if x not in self.schemas[schema] else x,
                        pl.String,
                    ),
                    pl.col("benchmark_id").map_elements(
                        lambda x: (
                            f"[{x}]"
                            if x is not None and x not in self.schemas[schema]
                            else x
                        ),
                        pl.String,
                    ),
                )
                return f"""Missing portfolios are enclosed in brackets.

{df.__repr__()}"""
            else:
                return df.__repr__()

    def validate_settings(self, settings: PortfolioHierarchySettings) -> None:
        schema = settings.portfolio_schema
        if isinstance(schema, str) and schema not in self.schemas:
            raise ValueError(
                f"Unknown schema: {schema}, available: {self.schemas.keys()}"
            )

        if isinstance(schema, str):
            available_portfolio_ids = set(self.schemas[schema])
            available_portfolio_ids_with_rest = {
                n.split(":")[0] + ":{REST}" for n in available_portfolio_ids if ":" in n
            } | available_portfolio_ids
            portfolio_ids = set(settings.portfolio_ids)
            benchmark_ids = {b for b in settings.benchmark_ids if b is not None}

            if not portfolio_ids.issubset(available_portfolio_ids_with_rest):
                raise ValueError(
                    f"There are unknown portfolio ids. "
                    f"Unknown: {portfolio_ids - available_portfolio_ids}"
                )

            if not benchmark_ids.issubset(available_portfolio_ids):
                raise ValueError(
                    f"There are unknown benchmark ids. "
                    f"Unknown: {benchmark_ids - available_portfolio_ids}"
                )



from bayesline.api._src.equity import settings as settings_tools
from bayesline.api._src.equity.api import AsyncBayeslineEquityApi, BayeslineEquityApi
from bayesline.api._src.equity.exposure import (
    AssetExposureApi,
    AsyncAssetExposureApi,
    AsyncBayeslineEquityExposureApi,
    BayeslineEquityExposureApi,
)
from bayesline.api._src.equity.exposure_settings import (
    ExposureSettings,
    ExposureSettingsMenu,
    HierarchyGroups,
    HierarchyLevel,
)
from bayesline.api._src.equity.ids import (
    AsyncBayeslineEquityIdApi,
    BayeslineEquityIdApi,
)
from bayesline.api._src.equity.modelconstruction import (
    AsyncBayeslineModelConstructionApi,
    AsyncModelConstructionEngineApi,
    BayeslineModelConstructionApi,
    ModelConstructionEngineApi,
)
from bayesline.api._src.equity.modelconstruction_settings import (
    ModelConstructionSettings,
    ModelConstructionSettingsMenu,
)
from bayesline.api._src.equity.portfolio import (
    AsyncBayeslineEquityPortfolioApi,
    AsyncBayeslineEquityPortfolioUploadApi,
    AsyncBayeslinePortfolioApi,
    AsyncBayeslinePortfolioUploadApi,
    AsyncPortfolioParser,
    BayeslineEquityPortfolioApi,
    BayeslineEquityPortfolioUploadApi,
    BayeslinePortfolioApi,
    BayeslinePortfolioUploadApi,
    PortfolioParser,
    PortfolioParserResult,
    PortfolioUploadError,
)
from bayesline.api._src.equity.portfolio_settings import (
    PortfolioOrganizerSettings,
    PortfolioOrganizerSettingsMenu,
    PortfolioSettings,
    PortfolioSettingsMenu,
    PortfolioUploadSettings,
    PortfolioUploadSettingsMenu,
)
from bayesline.api._src.equity.portfoliohierarchy import (
    AsyncBayeslinePortfolioHierarchyApi,
    AsyncPortfolioHierarchyApi,
    BayeslinePortfolioHierarchyApi,
    PortfolioHierarchyApi,
)
from bayesline.api._src.equity.portfoliohierarchy_settings import (
    PortfolioHierarchySettings,
    PortfolioHierarchySettingsMenu,
)
from bayesline.api._src.equity.portfolioreport import (
    AsyncBayeslinePortfolioReportApi,
    AsyncPortfolioReportAccessor,
    AsyncPortfolioReportApi,
    AsyncReportPersister,
    BayeslinePortfolioReportApi,
    IllegalPathError,
    PortfolioReportAccessor,
    PortfolioReportApi,
    ReportPersister,
)
from bayesline.api._src.equity.report_settings import (
    AssetForecastLossReportSettings,
    AssetHoldingsMeasureSettings,
    AssetHoldingsReportSettings,
    AssetHoldingsReportSettingsMenu,
    AssetStressTestMeasureSettings,
    AssetStressTestReportSettings,
    AssetStressTestReportSettingsMenu,
    AVaRMeasureSettings,
    BrinsonAttributionMeasureSettings,
    BrinsonAttributionReportSettings,
    BrinsonAttributionReportSettingsMenu,
    ConcreteReportSettings,
    ConcreteReportSettingsMenu,
    CumsumMeasureSettings,
    DrawdownMeasureSettings,
    ExposureMeasureSettings,
    ExposureReportSettings,
    ExposureReportSettingsMenu,
    FactorAttributionMeasureSettings,
    FactorAttributionReportSettings,
    FactorAttributionReportSettingsMenu,
    FactorCovarianceMeasureSettings,
    FactorCovarianceReportSettings,
    FactorCovarianceReportSettingsMenu,
    FactorForecastLossReportSettings,
    FactorIdioMeasureSettings,
    FactorIdioReportSettings,
    FactorIdioReportSettingsMenu,
    FactorMovingAverageMeasureSettings,
    FactorTSReportSettings,
    FactorVolForecastMeasureSettings,
    ForecastBacktestMeasureSettings,
    ForecastBacktestReportSettings,
    ForecastBacktestReportSettingsMenu,
    HoldingsMeasureSettings,
    HoldingsReportSettings,
    HoldingsReportSettingsMenu,
    HVaRMeasureSettings,
    IdiosyncraticVolatilityMeasureSettings,
    IdiosyncraticVolatilityReportSettings,
    IdiosyncraticVolatilityReportSettingsMenu,
    MeasureSettings,
    MovingAverageMeasureSettings,
    MovingAverageRSquaredMeasureSettings,
    PassThroughFactor2DMeasureSettings,
    PortfolioReportAccessorSettings,
    PortfolioStressTestMeasureSettings,
    PortfolioStressTestReportSettings,
    PortfolioStressTestReportSettingsMenu,
    ReportSettings,
    ReportSettingsMenu,
    RiskModelFitFactorReportSettings,
    RiskModelFitFactorReportSettingsMenu,
    RiskModelFitReportSettings,
    RiskModelPortfolioFitReportSettings,
    StyleCorrelationReportSettings,
    StyleCorrelationReportSettingsMenu,
    StyleIndustryExposureReportSettings,
    StyleIndustryExposureReportSettingsMenu,
    TimeSeriesBetaMeasureSettings,
    TimeSeriesVolatilityMeasureSettings,
    TimeSeriesXSigmaRhoMeasureSettings,
    TSBetaReportSettings,
    TSXSRReportSettings,
    TSXSRReportSettingsMenu,
    VaRReportSettings,
    VaRReportSettingsMenu,
    VolForecastMeasureSettings,
    XSigmaRhoMeasureSettings,
    XSRReportSettings,
    XSRReportSettingsMenu,
)
from bayesline.api._src.equity.riskmodels import (
    AsyncBayeslineFactorRiskModelsApi,
    AsyncFactorRiskEngineApi,
    AsyncFactorRiskModelApi,
    BayeslineFactorRiskModelsApi,
    FactorRiskEngineApi,
    FactorRiskModelApi,
    FactorType,
)
from bayesline.api._src.equity.riskmodels_settings import (
    FactorRiskModelSettings,
    FactorRiskModelSettingsMenu,
)
from bayesline.api._src.equity.universe import (
    AssetUniverseApi,
    AsyncAssetUniverseApi,
    AsyncBayeslineEquityUniverseApi,
    BayeslineEquityUniverseApi,
)
from bayesline.api._src.equity.universe_settings import (
    Hierarchy,
    IndustrySettings,
    MCapSettings,
    RegionSettings,
    UniverseSettings,
    UniverseSettingsMenu,
)

__all__ = [
    "BayeslineEquityApi",
    "AssetExposureApi",
    "AsyncAssetExposureApi",
    "BayeslineEquityUniverseApi",
    "BayeslineEquityExposureApi",
    "UniverseSettings",
    "UniverseSettingsMenu",
    "IndustrySettings",
    "RegionSettings",
    "MCapSettings",
    "Hierarchy",
    "AssetUniverseApi",
    "AsyncAssetUniverseApi",
    "HierarchyLevel",
    "HierarchyGroups",
    "MeasureSettings",
    "ExposureSettings",
    "ExposureSettingsMenu",
    "TimeSeriesVolatilityMeasureSettings",
    "TimeSeriesXSigmaRhoMeasureSettings",
    "CumsumMeasureSettings",
    "DrawdownMeasureSettings",
    "MovingAverageMeasureSettings",
    "FactorMovingAverageMeasureSettings",
    "RiskModelFitReportSettings",
    "RiskModelPortfolioFitReportSettings",
    "VolForecastMeasureSettings",
    "FactorVolForecastMeasureSettings",
    "FactorIdioMeasureSettings",
    "MovingAverageRSquaredMeasureSettings",
    "settings_tools",
    "ModelConstructionSettings",
    "ModelConstructionSettingsMenu",
    "ReportSettings",
    "ReportSettingsMenu",
    "ConcreteReportSettings",
    "ConcreteReportSettingsMenu",
    "TSXSRReportSettings",
    "TSXSRReportSettingsMenu",
    "XSRReportSettings",
    "XSRReportSettingsMenu",
    "VaRReportSettings",
    "VaRReportSettingsMenu",
    "ExposureMeasureSettings",
    "FactorTSReportSettings",
    "HVaRMeasureSettings",
    "AVaRMeasureSettings",
    "FactorForecastLossReportSettings",
    "AssetForecastLossReportSettings",
    "PortfolioReportApi",
    "AsyncPortfolioReportApi",
    "BayeslinePortfolioReportApi",
    "BayeslineModelConstructionApi",
    "FactorRiskEngineApi",
    "AsyncFactorRiskEngineApi",
    "FactorRiskModelApi",
    "AsyncFactorRiskModelApi",
    "FactorType",
    "FactorRiskModelSettings",
    "FactorRiskModelSettingsMenu",
    "BayeslineFactorRiskModelsApi",
    "ModelConstructionEngineApi",
    "AsyncModelConstructionEngineApi",
    "FactorIdioReportSettings",
    "RiskModelFitFactorReportSettings",
    "RiskModelFitFactorReportSettingsMenu",
    "PassThroughFactor2DMeasureSettings",
    "StyleCorrelationReportSettings",
    "StyleIndustryExposureReportSettings",
    "FactorIdioReportSettingsMenu",
    "StyleCorrelationReportSettingsMenu",
    "StyleIndustryExposureReportSettingsMenu",
    "AsyncBayeslineEquityApi",
    "AsyncBayeslineEquityExposureApi",
    "AsyncBayeslineEquityUniverseApi",
    "AsyncBayeslineFactorRiskModelsApi",
    "AsyncBayeslineModelConstructionApi",
    "AsyncBayeslinePortfolioReportApi",
    "XSigmaRhoMeasureSettings",
    "BayeslineEquityByodApi",
    "ByodApi",
    "AsyncByodApi",
    "ByodSettings",
    "ByodSettingsMenu",
    "InferAssetIdException",
    "ForecastBacktestReportSettings",
    "ForecastBacktestReportSettingsMenu",
    "ForecastBacktestMeasureSettings",
    "BrinsonAttributionMeasureSettings",
    "BrinsonAttributionReportSettings",
    "BrinsonAttributionReportSettingsMenu",
    "FactorAttributionMeasureSettings",
    "FactorAttributionReportSettings",
    "FactorAttributionReportSettingsMenu",
    "BayeslineEquityPortfolioApi",
    "AsyncBayeslineEquityPortfolioApi",
    "PortfolioStressTestMeasureSettings",
    "PortfolioStressTestReportSettings",
    "PortfolioStressTestReportSettingsMenu",
    "AssetStressTestMeasureSettings",
    "AssetStressTestReportSettings",
    "AssetStressTestReportSettingsMenu",
    "AssetHoldingsMeasureSettings",
    "AssetHoldingsReportSettings",
    "AssetHoldingsReportSettingsMenu",
    "PortfolioHierarchySettings",
    "PortfolioHierarchySettingsMenu",
    "BayeslinePortfolioHierarchyApi",
    "AsyncBayeslinePortfolioHierarchyApi",
    "AsyncPortfolioHierarchyApi",
    "PortfolioHierarchyApi",
    "HoldingsMeasureSettings",
    "HoldingsReportSettings",
    "HoldingsReportSettingsMenu",
    "BayeslineEquityIdApi",
    "AsyncBayeslineEquityIdApi",
    "ExposureMeasureSettings",
    "ExposureReportSettings",
    "ExposureReportSettingsMenu",
    "PortfolioSettings",
    "PortfolioSettingsMenu",
    "AsyncBayeslinePortfolioApi",
    "BayeslinePortfolioApi",
    "PortfolioUploadSettings",
    "PortfolioUploadSettingsMenu",
    "AsyncBayeslineEquityPortfolioUploadApi",
    "BayeslineEquityPortfolioUploadApi",
    "AsyncBayeslinePortfolioUploadApi",
    "BayeslinePortfolioUploadApi",
    "PortfolioParser",
    "PortfolioParserResult",
    "AsyncPortfolioParser",
    "PortfolioUploadError",
    "PortfolioOrganizerSettings",
    "PortfolioOrganizerSettingsMenu",
    "IdiosyncraticVolatilityMeasureSettings",
    "IdiosyncraticVolatilityReportSettings",
    "IdiosyncraticVolatilityReportSettingsMenu",
    "FactorCovarianceMeasureSettings",
    "FactorCovarianceReportSettings",
    "FactorCovarianceReportSettingsMenu",
    "IllegalPathError",
    "AsyncPortfolioReportAccessor",
    "PortfolioReportAccessorSettings",
    "PortfolioReportAccessor",
    "AsyncReportPersister",
    "ReportPersister",
    "TimeSeriesBetaMeasureSettings",
    "TSBetaReportSettings",
]



import abc
import datetime as dt
from typing import Literal

import polars as pl

from bayesline.api._src.equity.exposure_settings import (
    ExposureSettings,
    ExposureSettingsMenu,
)
from bayesline.api._src.equity.universe import AssetUniverseApi, AsyncAssetUniverseApi
from bayesline.api._src.equity.universe_settings import UniverseSettings
from bayesline.api._src.registry import AsyncRegistryBasedApi, RegistryBasedApi
from bayesline.api._src.types import DateLike, IdType

FactorType = Literal["Market", "Style", "Industry", "Region"]


class AssetExposureApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> ExposureSettings:
        """
        Returns
        -------
        The settings used to create these exposures.
        """
        ...

    @abc.abstractmethod
    def dates(
        self,
        universe: str | int | UniverseSettings | AssetUniverseApi,
        *,
        range_only: bool = False,
    ) -> list[dt.date]:
        """
        Parameters
        ----------
        universe: str | int | UniverseSettings | AssetUniverseApi
            The universe to use for the exposure calculation.
        range_only: bool, optional
            If True, returns the first and last date only.

        Returns
        -------
        list of all covered dates.
        """

    @abc.abstractmethod
    def coverage_stats(
        self,
        universe: str | int | UniverseSettings | AsyncAssetUniverseApi,
        *,
        id_type: IdType | None = None,
        by: Literal["date", "asset"] = "date",
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        universe: str | int | UniverseSettings | AssetUniverseApi
            The universe to use for the exposure calculation.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.
        by: str, optional
            The aggregation, either by date or by asset

        Returns
        -------
        pl.DataFrame
            A dataframe with date as the first column, where the remaining columns
            names are the styles and substyles (concatenated with a dot). The values
            are the counts of the underlying data before it was imputed.
        """

    @abc.abstractmethod
    def get(
        self,
        universe: str | int | UniverseSettings | AssetUniverseApi,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        universe: str | int | UniverseSettings | AssetUniverseApi
            The universe to use for the exposure calculation.
        start: DateLike, optional
            The start date of the universe to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range with a where the date is the first column
            and the asset id is the second column. The remaining columns are the
            individual styles.
        """
        ...


class AsyncAssetExposureApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> ExposureSettings:
        """
        Returns
        -------
        The settings used to create these exposures.
        """
        ...

    @abc.abstractmethod
    async def dates(
        self,
        universe: str | int | UniverseSettings | AsyncAssetUniverseApi,
        *,
        range_only: bool = False,
    ) -> list[dt.date]:
        """
        Parameters
        ----------
        universe: str | int | UniverseSettings | AssetUniverseApi
            The universe to use for the exposure calculation.
        range_only: bool, optional
            If True, returns the first and last date only.

        Returns
        -------
        list of all covered dates.
        """

    @abc.abstractmethod
    async def coverage_stats(
        self,
        universe: str | int | UniverseSettings | AsyncAssetUniverseApi,
        *,
        id_type: IdType | None = None,
        by: Literal["date", "asset"] = "date",
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        universe: str | int | UniverseSettings | AssetUniverseApi
            The universe to use for the exposure calculation.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.
        by: str, optional
            The aggregation, either by date or by asset

        Returns
        -------
        pl.DataFrame
            A dataframe with date as the first column, where the remaining columns
            names are the styles and substyles (concatenated with a dot). The values
            are the counts of the underlying data before it was imputed.
        """

    @abc.abstractmethod
    async def get(
        self,
        universe: str | int | UniverseSettings | AsyncAssetUniverseApi,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        universe: str | int | UniverseSettings | AssetUniverseApi
            The universe to use for the exposure calculation.
        start: DateLike, optional
            The start date of the universe to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range with a where the date is the first column
            and the asset id is the second column. The remaining columns are the
            individual styles.
        """
        ...


class BayeslineEquityExposureApi(
    RegistryBasedApi[ExposureSettings, ExposureSettingsMenu, AssetExposureApi],
): ...


class AsyncBayeslineEquityExposureApi(
    AsyncRegistryBasedApi[
        ExposureSettings, ExposureSettingsMenu, AsyncAssetExposureApi
    ],
): ...



__version__ = "0.0.12post"

from bayesline.api._src.api import AsyncBayeslineApi, BayeslineApi
from bayesline.api._src.permissions import AsyncUserPermissionsApi, UserPermissionsApi
from bayesline.api._src.registry import (
    AsyncInMemorySettingsRegistry,
    AsyncReadOnlyRegistry,
    AsyncRegistry,
    AsyncSettingsRegistry,
    AsyncSettingsResolver,
    InvalidSettingsError,
    Mode,
    RawSettings,
    ReadOnlyRegistry,
    Registry,
    Settings,
    SettingsMenu,
    SettingsMenuType,
    SettingsMetaData,
    SettingsRegistry,
    SettingsType,
    SettingsTypeMetaData,
)

__all__ = [
    "SettingsType",
    "SettingsMenuType",
    "types",
    "SettingsMenu",
    "BayeslineApi",
    "AsyncBayeslineApi",
    "SettingsRegistry",
    "AsyncSettingsRegistry",
    "AsyncInMemorySettingsRegistry",
    "Registry",
    "AsyncRegistry",
    "SettingsMetaData",
    "UserPermissionsApi",
    "AsyncUserPermissionsApi",
    "DynamicAnalyticsMessage",
    "DynamicAnalyticsResponse",
    "ModelDynamicAnalyticsResponse",
    "DynamicAnalyticsSession",
    "DynamicAnalyticsService",
    "DynamicAnalyticsTableColumnDefinition",
    "DynamicAnalyticsTableSchema",
    "DynamicAnalyticsTableMessage",
    "DynamicAnalyticsOptionsMessage",
    "BaseDynamicAnalyticsResponse",
    "DynamicAnalyticsOptionsMessageFormat",
    "DynamicAnalyticsMessageFormat",
    "DynamicAnalyticsSessionModel",
    "DynamicAnalyticsSessionRecorder",
    "Settings",
    "SettingsTypeMetaData",
    "Mode",
    "RawSettings",
    "InvalidSettingsError",
    "AsyncSettingsResolver",
    "ReadOnlyRegistry",
    "AsyncReadOnlyRegistry",
]



from bayesline.api._src.types import (
    DateLike,
    DNFFilterExpression,
    DNFFilterExpressions,
    IdType,
    to_date,
    to_date_string,
    to_maybe_date,
    to_maybe_date_string,
)

__all__ = [
    "IdType",
    "DateLike",
    "DNFFilterExpression",
    "DNFFilterExpressions",
    "to_date",
    "to_date_string",
    "to_maybe_date",
    "to_maybe_date_string",
]



import os
from typing import Annotated

from pydantic import Field

from bayesline.api._src.equity.exposure_settings import ExposureSettings
from bayesline.api._src.equity.modelconstruction_settings import (
    ModelConstructionSettings,
)
from bayesline.api._src.equity.universe_settings import UniverseSettings
from bayesline.api._src.registry import Settings, SettingsMenu, SettingsTypeMetaData


class FactorRiskModelSettings(Settings):
    """
    Defines all settings needed to build a factor risk model.
    """

    @classmethod
    def default(cls) -> "FactorRiskModelSettings":
        return cls(
            universe=UniverseSettings.default(),
            exposures=ExposureSettings.default(),
            modelconstruction=ModelConstructionSettings.default(),
        )

    universe: Annotated[
        str | int | UniverseSettings,
        Field(
            description="The universe to build the factor risk model on.",
            default_factory=UniverseSettings.default,
        ),
        SettingsTypeMetaData[str | int | UniverseSettings](references=UniverseSettings),
    ]

    exposures: Annotated[
        str | int | ExposureSettings,
        Field(
            description="The exposures to build the factor risk model on.",
            default_factory=ExposureSettings.default,
        ),
        SettingsTypeMetaData[str | int | ExposureSettings](references=ExposureSettings),
    ]

    modelconstruction: Annotated[
        str | int | ModelConstructionSettings,
        Field(
            description="The model construction settings to use for the factor risk model.",
            default_factory=ModelConstructionSettings.default,
        ),
        SettingsTypeMetaData[str | int | ModelConstructionSettings](
            references=ModelConstructionSettings
        ),
    ]


class FactorRiskModelSettingsMenu(SettingsMenu, frozen=True, extra="forbid"):
    """
    Defines available settings to build a factor risk model.
    """

    def describe(self, settings: FactorRiskModelSettings | None = None) -> str:
        if settings:
            result = [
                "Universe: " + str(settings.universe),
                "Exposures: " + str(settings.exposures),
                "Model Construction: " + str(settings.modelconstruction),
            ]
            return os.linesep.join(result)
        else:
            return "This settings menu has no description."

    def validate_settings(self, settings: FactorRiskModelSettings) -> None:
        if not isinstance(settings.exposures, ExposureSettings) or not isinstance(
            settings.modelconstruction, ModelConstructionSettings
        ):
            return

        known_factors = set(settings.modelconstruction.known_factors)
        other_factors = set(settings.exposures.other)
        missing_known_exposures = known_factors - other_factors
        if missing_known_exposures:
            raise ValueError(
                f"Invalid known factor returns: {', '.join(missing_known_exposures)}"
            )



import os
from collections import Counter, defaultdict
from collections.abc import Callable, Iterable, Mapping
from typing import TYPE_CHECKING, TypeAlias

from typing_extensions import TypeAliasType

if TYPE_CHECKING:
    Hierarchy: TypeAlias = "list[str] | Mapping[str, Hierarchy]"
else:
    Hierarchy = TypeAliasType("Hierarchy", list[str] | Mapping[str, "Hierarchy"])


def check_unique_hierarchy(
    v: Mapping[str, Mapping[str, Hierarchy]],
) -> Mapping[str, Mapping[str, Hierarchy]]:
    errors = []
    for hierarchy_name, hierarchy in v.items():
        flat_hierarchy = flatten(hierarchy)
        dupes = [item for item, count in Counter(flat_hierarchy).items() if count > 1]

        if dupes:
            errors.append(
                f"""
                There are duplicate elements in hierarchy {hierarchy_name}:
                {', '.join(dupes)}
                """,
            )
    if errors:
        raise ValueError(os.linesep.join(errors))
    else:
        return v


def check_nonempty_hierarchy(
    v: Mapping[str, Mapping[str, Hierarchy]],
) -> Mapping[str, Mapping[str, Hierarchy]]:
    errors = []
    if len(v) == 0:
        errors.append("Hierarchy cannot be empty.")
    else:
        for hierarchy_name, hierarchy in v.items():
            if len(hierarchy) == 0:
                errors.append(f"Hierarchy {hierarchy_name} cannot be empty.")

    if errors:
        raise ValueError(os.linesep.join(errors))
    else:
        return v


def check_all_codes_have_labels(
    hierarchies: Mapping[str, Hierarchy],
    labels: Mapping[str, Mapping[str, str]],
) -> list[str]:
    missing_hierarchies = []
    missing_codes = defaultdict(list)
    extra_codes = defaultdict(list)

    for hierarchy_name, hierarchy in hierarchies.items():
        if hierarchy_name not in labels:
            missing_hierarchies.append(hierarchy_name)
            break

        all_codes = set(flatten(hierarchy))
        all_label_codes = set(labels[hierarchy_name].keys())
        missing_codes[hierarchy_name].extend(list(all_codes - all_label_codes))
        extra_codes[hierarchy_name].extend(list(all_label_codes - all_codes))

    error_messages = []
    if missing_hierarchies:
        missing_str = ", ".join(missing_hierarchies)
        error_messages.append(
            f"The following hierarchies do not have mappings: {missing_str}",
        )

    for hierarchy_name, codes in missing_codes.items():
        if codes:
            missing_str = ", ".join(codes)
            error_messages.append(
                f"Hierarchies {hierarchy_name} has missing mappings: {missing_str}",
            )

    for hierarchy_name, codes in extra_codes.items():
        if codes:
            xtr_str = ", ".join(codes)
            error_messages.append(
                f"Hierarchies {hierarchy_name} has superfluous mappings: {xtr_str}",
            )

    return error_messages


def flatten(
    hierarchy: Hierarchy,
    only_leaves: bool = False,
) -> list[str]:
    if isinstance(hierarchy, list):
        return hierarchy

    items = []
    for k, v in hierarchy.items():
        if not only_leaves:
            items.append(k)
        if isinstance(v, dict):
            items.extend(flatten(v, only_leaves=only_leaves))
        else:
            items.extend(v)
    return items


def validate_hierarchy_schema(
    available_schemas: Iterable[str],
    hierarchy_schema: str,
) -> None:
    if hierarchy_schema not in available_schemas:
        raise ValueError(
            f"""
            Schema {hierarchy_schema} does not exist.
            Only {', '.join(available_schemas)} exist.
            """,
        )


def map_hierarchy(
    hierarchy: Mapping[str, Hierarchy],
    fn: Callable[[str], str],
) -> Mapping[str, Hierarchy]:
    return {
        fn(k): (map_hierarchy(v, fn) if not isinstance(v, list) else [fn(e) for e in v])
        for k, v in hierarchy.items()
    }


def filter_leaves(
    hierarchy: Mapping[str, Hierarchy],
    fn: Callable[[str], bool],
) -> Mapping[str, Hierarchy]:
    result: Mapping[str, Hierarchy] = {}
    for k, v in hierarchy.items():
        if isinstance(v, dict):
            survivors1 = filter_leaves(v, fn)
            if survivors1:
                result[k] = survivors1  # type: ignore
        else:
            survivors2 = [e for e in v if fn(e)]
            if survivors2:
                result[k] = survivors2  # type: ignore
    return result


def find_in_hierarchy(
    hierarchy: Mapping[str, Hierarchy],
    needle: str,
    *,
    return_depth: bool = False,
    depth: int = 0,
) -> Hierarchy | int | None:
    for k, v in hierarchy.items():
        if k == needle:
            return {k: v} if not return_depth else depth + 1
        if isinstance(v, list):
            if needle in v:
                return [needle] if not return_depth else depth + 2
        elif isinstance(v, dict):
            match = find_in_hierarchy(
                v,
                needle,
                return_depth=return_depth,
                depth=depth + 1,
            )
            if match:
                return match if not return_depth else match

    if depth == 0:
        return None if not return_depth else -1
    else:
        return []


def trim_to_depth(hierarchy: Hierarchy, depth: int) -> Hierarchy:
    if depth < 0:
        raise AssertionError(f"Depth must be >= 0., {depth}")
    if depth == 1:
        return list(hierarchy.keys()) if isinstance(hierarchy, dict) else hierarchy
    elif isinstance(hierarchy, list):
        return hierarchy
    else:
        return {k: trim_to_depth(v, depth - 1) for k, v in hierarchy.items()}


def filter_labels_to_hierarchy(
    hierarchy: Mapping[str, Hierarchy],
    labels: Mapping[str, str],
) -> Mapping[str, str]:
    return {k: v for k, v in labels.items() if k in flatten(hierarchy)}


def get_depth(hierarchy: Mapping[str, Hierarchy] | list[str]) -> int:
    if isinstance(hierarchy, list):
        d = 0
    else:
        depths = []
        for value in hierarchy.values():
            depths.append(get_depth(value))
        d = max(depths)
    return d + 1


def check_all_leafcodes_exist(
    hierarchy: Mapping[str, Hierarchy], leaf_codes: set[str]
) -> None:
    missing_codes = set(flatten(hierarchy, only_leaves=True)) - leaf_codes
    if missing_codes:
        raise ValueError(f"There are missing codes: {', '.join(missing_codes)}")


# TODO: move somewhere else
def unnest(hierarchy: Mapping[str, Hierarchy] | list[str]) -> list[list[str]]:
    """Convert a nested hierarchy to a list of lists."""
    if isinstance(hierarchy, list):
        return [[v] for v in hierarchy]
    elif isinstance(hierarchy, dict):
        return sum([[[k, *v] for v in unnest(h)] for k, h in hierarchy.items()], [])
    else:
        raise ValueError("Hierarchy must be either a list or a dictionary.")


def prune_leaves(
    hierarchy: Mapping[str, Hierarchy],
) -> Mapping[str, Hierarchy]:
    return {
        k: prune_leaves(v) if not isinstance(v, list) else []
        for k, v in hierarchy.items()
    }



import abc

from bayesline.api._src.equity.exposure import (
    AsyncBayeslineEquityExposureApi,
    BayeslineEquityExposureApi,
)
from bayesline.api._src.equity.ids import (
    AsyncBayeslineEquityIdApi,
    BayeslineEquityIdApi,
)
from bayesline.api._src.equity.modelconstruction import (
    AsyncBayeslineModelConstructionApi,
    BayeslineModelConstructionApi,
)
from bayesline.api._src.equity.portfolio import (
    AsyncBayeslineEquityPortfolioApi,
    BayeslineEquityPortfolioApi,
)
from bayesline.api._src.equity.portfoliohierarchy import (
    AsyncBayeslinePortfolioHierarchyApi,
    BayeslinePortfolioHierarchyApi,
)
from bayesline.api._src.equity.portfolioreport import (
    AsyncBayeslinePortfolioReportApi,
    BayeslinePortfolioReportApi,
)
from bayesline.api._src.equity.riskmodels import (
    AsyncBayeslineFactorRiskModelsApi,
    BayeslineFactorRiskModelsApi,
)
from bayesline.api._src.equity.universe import (
    AsyncBayeslineEquityUniverseApi,
    BayeslineEquityUniverseApi,
)


class BayeslineEquityApi(abc.ABC):

    @property
    @abc.abstractmethod
    def ids(self) -> BayeslineEquityIdApi: ...

    @property
    @abc.abstractmethod
    def universes(self) -> BayeslineEquityUniverseApi: ...

    @property
    @abc.abstractmethod
    def exposures(self) -> BayeslineEquityExposureApi: ...

    @property
    @abc.abstractmethod
    def modelconstruction(self) -> BayeslineModelConstructionApi: ...

    @property
    @abc.abstractmethod
    def riskmodels(self) -> BayeslineFactorRiskModelsApi: ...

    @property
    @abc.abstractmethod
    def portfoliohierarchy(self) -> BayeslinePortfolioHierarchyApi: ...

    @property
    @abc.abstractmethod
    def portfolioreport(self) -> BayeslinePortfolioReportApi: ...

    @property
    @abc.abstractmethod
    def portfolios(self) -> BayeslineEquityPortfolioApi: ...


class AsyncBayeslineEquityApi(abc.ABC):

    @property
    @abc.abstractmethod
    def ids(self) -> AsyncBayeslineEquityIdApi: ...

    @property
    @abc.abstractmethod
    def universes(self) -> AsyncBayeslineEquityUniverseApi: ...

    @property
    @abc.abstractmethod
    def exposures(self) -> AsyncBayeslineEquityExposureApi: ...

    @property
    @abc.abstractmethod
    def modelconstruction(self) -> AsyncBayeslineModelConstructionApi: ...

    @property
    @abc.abstractmethod
    def riskmodels(self) -> AsyncBayeslineFactorRiskModelsApi: ...

    @property
    @abc.abstractmethod
    def portfoliohierarchy(self) -> AsyncBayeslinePortfolioHierarchyApi: ...

    @property
    @abc.abstractmethod
    def portfolioreport(self) -> AsyncBayeslinePortfolioReportApi: ...

    @property
    @abc.abstractmethod
    def portfolios(self) -> AsyncBayeslineEquityPortfolioApi: ...



import abc
import datetime as dt

import polars as pl

from bayesline.api._src.equity.portfoliohierarchy_settings import (
    PortfolioHierarchySettings,
    PortfolioHierarchySettingsMenu,
)
from bayesline.api._src.registry import AsyncRegistryBasedApi, RegistryBasedApi
from bayesline.api._src.types import DateLike, IdType


class PortfolioHierarchyApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> PortfolioHierarchySettings:
        """
        Returns
        -------
        The settings used to create this hierarchy.
        """
        ...

    @abc.abstractmethod
    def get_id_types(self) -> dict[str, list[IdType]]:
        """
        Returns
        -------
        dict[str, list[IdType]]:
            The available ID types that at least a portion of assets can be mapped
            to for each portfolio.
            If a portfolio has a benchmark then the available id types are those
            that are available for both the portfolio and the benchmark.
        """

    @abc.abstractmethod
    def get_dates(self, *, collapse: bool = False) -> dict[str, list[dt.date]]:
        """
        Parameters
        ----------
        collapse: bool, optional
            If True, will calculate aggregations `any` and `all`, indicating
            of for a given date, any (or all) portfolios have holdings.

        Returns
        -------
        A dict of portfolio-id to dates for which this hierarchy can be produced.
        For a given portfolio and date, the hierarchy can be produced if
        the portfolio has holdings for that date. If a benchmark is given then
        this benchmark also must have holdings for the given date.
        """

    @abc.abstractmethod
    def get(
        self,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame: ...


class AsyncPortfolioHierarchyApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> PortfolioHierarchySettings:
        """
        Returns
        -------
        The settings used to create this hierarchy.
        """
        ...

    @abc.abstractmethod
    async def get_id_types(self) -> dict[str, list[IdType]]:
        """
        Returns
        -------
        dict[str, list[IdType]]:
            The available ID types that at least a portion of assets can be mapped
            to for each portfolio.
            If a portfolio has a benchmark then the available id types are those
            that are available for both the portfolio and the benchmark.
        """

    @abc.abstractmethod
    async def get_dates(self, *, collapse: bool = False) -> dict[str, list[dt.date]]:
        """
        Parameters
        ----------
        collapse: bool, optional
            If True, will calculate aggregations `any` and `all`, indicating
            of for a given date, any (or all) portfolios have holdings.

        Returns
        -------
        A dict of portfolio-id to dates for which this hierarchy can be produced.
        For a given portfolio and date, the hierarchy can be produced if
        the portfolio has holdings for that date. If a benchmark is given then
        this benchmark also must have holdings for the given date.
        """

    @abc.abstractmethod
    async def get(
        self,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame: ...


class BayeslinePortfolioHierarchyApi(
    RegistryBasedApi[
        PortfolioHierarchySettings,
        PortfolioHierarchySettingsMenu,
        PortfolioHierarchyApi,
    ]
): ...


class AsyncBayeslinePortfolioHierarchyApi(
    AsyncRegistryBasedApi[
        PortfolioHierarchySettings,
        PortfolioHierarchySettingsMenu,
        AsyncPortfolioHierarchyApi,
    ]
): ...
